#if 0
#include "vld.h"
#endif

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <set>

#ifdef _WIN32
#define NOMINMAX
#include "windows.h"
#endif

#define AG_AVLTREE_TESTING
#include <ag/avltree/avl_tree.h>

#define BOOST_TEST_MODULE avl_tree_test

#include <boost/test/unit_test.hpp>
#include <boost/test/test_tools.hpp>

using boost::unit_test::test_suite;

#define FUNCTION_HEADER std::cout << "---" << __FUNCTION__ << "---" << std::endl;


BOOST_AUTO_TEST_SUITE(avl_basics)


/**
* Tests srr
*/
BOOST_AUTO_TEST_CASE(test_avl_btree_srr) {
    FUNCTION_HEADER;    
    using namespace ag::avltree;
    AvlTree<int,int> my_tree;
    my_tree.setBalancing(true);

    my_tree.put(5,5);
    my_tree.put(3,3);
    my_tree.put(2,2); // should trigger balancing

    // my_tree.dumpMap_("\\tmp\\Balancing.dot");

    BOOST_CHECK(my_tree.get(2) == 2); 
    BOOST_CHECK(my_tree.get(3) == 3); 
    BOOST_CHECK(my_tree.get(5) == 5); 

    // test that the balancing info is still ok
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node) == 0);
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node->right_node) == 0);
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node->left_node) == 0);
    BOOST_CHECK(my_tree.root_node->key == 3);
    BOOST_CHECK(my_tree.root_node->right_node->key == 5);
    BOOST_CHECK(my_tree.root_node->left_node->key == 2);
}


/**
* Tests slr
*/
BOOST_AUTO_TEST_CASE(test_avl_btree_slr) {
    FUNCTION_HEADER;    
    using namespace ag::avltree;
    AvlTree<int,int> my_tree;
    my_tree.setBalancing(true);

    my_tree.put(3,3);
    my_tree.put(5,5);
    my_tree.put(7,7); // should trigger balancing: slr

    my_tree.dumpMap_("\\tmp\\Balancing.dot");

    BOOST_CHECK(my_tree.get(5) == 5); 
    BOOST_CHECK(my_tree.get(3) == 3); 
    BOOST_CHECK(my_tree.get(7) == 7); 

    // test that the balancing info is still ok
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node) == 0);
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node->right_node) == 0);
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node->left_node) == 0);
    BOOST_CHECK(my_tree.root_node->key == 5);
    BOOST_CHECK(my_tree.root_node->right_node->key == 7);
    BOOST_CHECK(my_tree.root_node->left_node->key == 3);
}


/**
* Tests doublerotate (left/right)
*/
BOOST_AUTO_TEST_CASE(test_avl_btree_doublerotate_leftright) {
    FUNCTION_HEADER;    
    using namespace ag::avltree;
    AvlTree<int,int> my_tree;
    my_tree.setBalancing(true);

    my_tree.put(5,5);
    my_tree.put(3,3);
    my_tree.put(4,4); 

    BOOST_CHECK(my_tree.get(5) == 5); 
    BOOST_CHECK(my_tree.get(3) == 3); 
    BOOST_CHECK(my_tree.get(4) == 4); 

    // test that the balancing info is still ok
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node) == 0);
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node->right_node) == 0);
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node->left_node) == 0);
    BOOST_CHECK(my_tree.root_node->key == 4);
    BOOST_CHECK(my_tree.root_node->right_node->key == 5);
    BOOST_CHECK(my_tree.root_node->left_node->key == 3);
}


/**
* Tests drotate (right/left)
*/
BOOST_AUTO_TEST_CASE(test_avl_btree_doublerotate_rightleft) {
    FUNCTION_HEADER;    
    using namespace ag::avltree;
    AvlTree<int,int> my_tree;
    my_tree.setBalancing(true);

    my_tree.put(3,3);
    my_tree.put(5,5);    
    my_tree.put(4,4); 

    BOOST_CHECK(my_tree.get(5) == 5); 
    BOOST_CHECK(my_tree.get(3) == 3); 
    BOOST_CHECK(my_tree.get(4) == 4); 

    // test that the balancing info is still ok
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node) == 0);
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node->right_node) == 0);
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node->left_node) == 0);
    BOOST_CHECK(my_tree.root_node->key == 4);
    BOOST_CHECK(my_tree.root_node->right_node->key == 5);
    BOOST_CHECK(my_tree.root_node->left_node->key == 3);
}


/**
* Tests btree deletion
*/
BOOST_AUTO_TEST_CASE(test_avl_btree_deletion) {
    FUNCTION_HEADER;    
    using namespace ag::avltree;
    AvlTree<int,int> my_tree;
    my_tree.setBalancing(true);

    // one child per leaf, case 1
    my_tree.put(1,1);
    my_tree.put(2,2);
    my_tree.put(3,3);        
    my_tree.put(4,4);    
    my_tree.put(22,22);    
    my_tree.put(13,13);    
    my_tree.remove(1);    
    my_tree.remove(22);
    my_tree.dumpMap_("C:\\tmp\\afterdel2.dot");

    // one child per leaf, case 2
    my_tree.reset();
    my_tree.put(5,5);
    my_tree.put(1,1);
    my_tree.put(4,4);
    my_tree.put(3,3);            
    my_tree.remove(1);
    my_tree.dumpMap_("C:\\tmp\\afterdel3b.dot");

    // two children per leaf, leaf is directly next to deletion node
    my_tree.reset();
    my_tree.put(1,1);
    my_tree.put(2,2);
    my_tree.put(3,3);        
    my_tree.put(4,4);
    my_tree.put(5,5);
    my_tree.dumpMap_("C:\\tmp\\afterdel4a.dot");
    my_tree.remove(4);
    my_tree.dumpMap_("C:\\tmp\\afterdel4b.dot");

    // two children per leaf, leaf is far away
    my_tree.reset();
    my_tree.put(1,1);
    my_tree.put(2,2);
    my_tree.put(3,3);        
    my_tree.put(4,4);
    my_tree.put(5,5);
    my_tree.put(6,6);
    my_tree.put(7,7);
    my_tree.put(8,8);
    my_tree.put(9,9);
    my_tree.dumpMap_("C:\\tmp\\afterdel5a.dot");
    my_tree.remove(6);
    my_tree.dumpMap_("C:\\tmp\\afterdel5b.dot");
    my_tree.remove(7);
    my_tree.dumpMap_("C:\\tmp\\afterdel5c.dot");
    my_tree.remove(8);
    my_tree.dumpMap_("C:\\tmp\\afterdel5d.dot");
    // delete the root :-)
    my_tree.remove(4);
    my_tree.dumpMap_("C:\\tmp\\afterdel5e.dot");
}


/**
* Tests random btree deletion
*/
BOOST_AUTO_TEST_CASE(test_avl_btree_random_deletions) {
    FUNCTION_HEADER;    
    using namespace ag::avltree;
    AvlTree<int,int> my_tree;
    my_tree.setBalancing(true);

    std::set<int> nums;
    std::map<int,int> map_check;
    std::set<int>::iterator it,end;

    int RUNS = 500;

    for (int x=0; x<RUNS; x++) {         
        int random_number = std::rand() * std::rand() * std::rand();
        nums.insert(random_number);
    }
    for (it=nums.begin(), end=nums.end(); it != end; ++it) {
        my_tree.put(*it ,*it);    
    }

    for (it=nums.begin(), end=nums.end(); it != end; ++it) {
        BOOST_CHECK(my_tree.get(*it) == *it);    
    }    

    int root = my_tree.root_node->key;
    my_tree.remove(root);
    nums.erase(root);
    for (it=nums.begin(), end=nums.end(); it != end; ++it) {
        BOOST_CHECK(my_tree.get(*it) == *it);    
    } 

    int MAXDELS = RUNS;
    int z=0;
    for (it=nums.begin(), end=nums.end(), z=0; it != end, z<MAXDELS; ++it,z++) {           
        std::cout << z << std::endl;     
        if (z == RUNS/2) {
            my_tree.dumpMap_("\\tmp\\BalancingHalf.dot");
        }   
        my_tree.remove(*it);    
    }
}


/**
* Tests insertion with random values 
*/
BOOST_AUTO_TEST_CASE(test_avl_btree_balancing_insert) {
    FUNCTION_HEADER;    
    using namespace ag::avltree;
    AvlTree<int,int> my_tree;  

    std::set<int> nums;
    std::map<int,int> map_check;
    std::set<int>::iterator it,end;

    int RUNS = 1000;

    std::cout << ::GetTickCount() << " Generating random numbers" << std::endl;
    for (int x=0; x<RUNS; x++) {         
        int random_number = std::rand() * std::rand() * std::rand();
        nums.insert(random_number);
    }
    std::cout << std::endl;
    std::cout << ::GetTickCount() << " Generating random numbers done, size:" << nums.size() << std::endl;


    std::cout << ::GetTickCount() << " Inserting" << std::endl;    
    for (it=nums.begin(), end=nums.end(); it != end; ++it) {
        my_tree.put(*it ,*it);    
    }
    std::cout << ::GetTickCount() << " Inserting done" << std::endl;


    std::cout << ::GetTickCount() << " Inserting for std::map" << std::endl;       
    for (it=nums.begin(), end=nums.end(); it != end; ++it) {
        map_check[*it] = *it;    
    }   
    std::cout << ::GetTickCount() << " Inserting for std::map done" << std::endl;

    //    std::stringstream fileNameStream; 
    //    fileNameStream << "\\tmp\\CheckTree-" << 100 << ".dot";
    //    my_tree.dumpMap_(fileNameStream.str());

    std::cout << ::GetTickCount() << " Retrieving" << std::endl;
    int start_tick = ::GetTickCount();
    for (it = nums.begin(), end = nums.end(); it != end; ++it) {
        BOOST_CHECK(my_tree.get(*it) == *it);
    }
    int end_tick = GetTickCount() - start_tick;
    std::cout << ::GetTickCount() << " Retrieving done" << std::endl;
    std::cout << "Msec: " << end_tick << std::endl;


    std::cout << ::GetTickCount() << " Retrieving for std::map" << std::endl;
    start_tick = ::GetTickCount();
    for (it = nums.begin(), end = nums.end(); it != end; ++it) {
        BOOST_CHECK(map_check.find(*it) != map_check.end());
    }
    end_tick = GetTickCount() - start_tick;
    std::cout << ::GetTickCount() << " Retrieving for std::map done" << std::endl;
    std::cout << "Msec: " << end_tick << std::endl;    
}          


/**
* Tests basics of the avl tree
*/
BOOST_AUTO_TEST_CASE(test_avl_btree_basics) {
    FUNCTION_HEADER;    
    using namespace ag::avltree;
    AvlTree<int,int> my_tree;
    my_tree.setBalancing(false); // disables balancing

    // first element 
    my_tree.put(5,66);
    BOOST_CHECK(my_tree.get(5) == 66);     

    // add left
    my_tree.put(1,12);
    BOOST_CHECK(my_tree.get(1) == 12); 

    // add right 
    my_tree.put(9,66);
    BOOST_CHECK(my_tree.get(9) == 66); 

    // size
    BOOST_CHECK(my_tree.getSize() == 3); 

    // query for nothing existing
    BOOST_CHECK(my_tree.get(777) == 0); 

    // check internal consistency
    AvlTree<int,int>::AvlNode* resultNode;
    resultNode = my_tree.getNode(9);
    BOOST_CHECK(resultNode->parent_node->left_max_depth == 1);
    BOOST_CHECK(resultNode->parent_node->right_max_depth == 1);
    BOOST_CHECK(resultNode->parent_node->left_node->left_max_depth == 0);
    BOOST_CHECK(resultNode->parent_node->left_node->right_max_depth == 0);

    // put another one to the far 
    my_tree.put(12,12);
    BOOST_CHECK(my_tree.get(12) == 12); 
    resultNode = my_tree.getNode(12);
    BOOST_CHECK(resultNode->parent_node->left_max_depth == 0);
    BOOST_CHECK(resultNode->parent_node->right_max_depth == 1);
    BOOST_CHECK(resultNode->parent_node->parent_node->right_max_depth == 2);
    BOOST_CHECK(resultNode->parent_node->parent_node->parent_node == 0);

    // put another one 
    my_tree.put(10,10);
    BOOST_CHECK(my_tree.get(10) == 10); 
    resultNode = my_tree.getNode(10);
    BOOST_CHECK(resultNode->parent_node->val == 12);
    BOOST_CHECK(resultNode->parent_node->val == 12);
    BOOST_CHECK(resultNode->parent_node->parent_node->parent_node->right_max_depth == 3);
    BOOST_CHECK(resultNode->parent_node->parent_node->parent_node->parent_node == 0);

    // test balance information
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node) == 2);
    my_tree.put(3,3);
    BOOST_CHECK(my_tree.get(3) == 3); 
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node) == 1);
    my_tree.put(2,2);
    BOOST_CHECK(my_tree.get(2) == 2); 
    BOOST_CHECK(my_tree.getBalance_(my_tree.root_node) == 0);

    // test total removal
    my_tree.reset();
    BOOST_CHECK(my_tree.getSize() == 0);
    BOOST_CHECK(my_tree.get(10) == 0);     
}

BOOST_AUTO_TEST_SUITE_END()
